
// This file is part of lmdb-store
// Copyright (c) 2013-2017 Timur Kristóf
// Licensed to you under the terms of the MIT license
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#ifndef NODE_LMDB_H
#define NODE_LMDB_H

#include <vector>
#include <algorithm>
#include <v8.h>
#include <node.h>
#include <node_buffer.h>
#include <nan.h>
#include <uv.h>
#include "lmdb.h"
#include "lz4.h"
#ifdef MDB_RPAGE_CACHE
#include "chacha8.h"
#endif
#if ENABLE_FAST_API && NODE_VERSION_AT_LEAST(16,6,0)
#include "v8-fast-api-calls.h"
#endif

using namespace v8;
using namespace node;

class Logging {
  public:
    static int debugLogging;
    static int initLogging();
};

enum class NodeLmdbKeyType {

    // Invalid key (used internally by node-lmdb)
    InvalidKey = -1,
    
    // Default key (used internally by node-lmdb)
    DefaultKey = 0,

    // UCS-2/UTF-16 with zero terminator - Appears to V8 as string
    StringKey = 1,
    
    // LMDB fixed size integer key with 32 bit keys - Appearts to V8 as an Uint32
    Uint32Key = 2,
    
    // LMDB default key format - Appears to V8 as node::Buffer
    BinaryKey = 3,

};
enum class KeyCreation {
    Reset = 0,
    Continue = 1,
    InArray = 2,
};
const int THEAD_MEMORY_THRESHOLD = 4000;

class TxnWrap;
class DbiWrap;
class EnvWrap;
class CursorWrap;
class Compression;
class KeySpace;

// Exports misc stuff to the module
void setupExportMisc(Local<Object> exports);

// Helper callback
typedef void (*argtokey_callback_t)(MDB_val &key);

void consoleLog(Local<Value> val);
void consoleLog(const char *msg);
void consoleLogN(int n);
void setFlagFromValue(int *flags, int flag, const char *name, bool defaultValue, Local<Object> options);
void writeValueToEntry(const Local<Value> &str, MDB_val *val);
argtokey_callback_t argToKey(const Local<Value> &val, MDB_val &key, NodeLmdbKeyType keyType, bool &isValid);
size_t valueToKey(const Local<Value> &jsKey, uint8_t* targetBytes, size_t remainingBytes, bool inArray, bool throwErrors);
bool valueToMDBKey(const Local<Value> &key, MDB_val &val, KeySpace &keySpace);

NodeLmdbKeyType inferAndValidateKeyType(const Local<Value> &key, const Local<Value> &options, NodeLmdbKeyType dbiKeyType, bool &isValid);
NodeLmdbKeyType inferKeyType(const Local<Value> &val);
NodeLmdbKeyType keyTypeFromOptions(const Local<Value> &val, NodeLmdbKeyType defaultKeyType = NodeLmdbKeyType::DefaultKey);
Local<Value> keyToHandle(MDB_val &key, NodeLmdbKeyType keyType);
bool getVersionAndUncompress(MDB_val &data, DbiWrap* dw);
int compare32LE(const MDB_val *a, const MDB_val *b);
NAN_METHOD(getLastVersion);
NAN_METHOD(setLastVersion);
NAN_METHOD(lmdbError);
NAN_METHOD(bufferToKeyValue);
NAN_METHOD(keyValueToBuffer);
//NAN_METHOD(getBufferForAddress);
NAN_METHOD(getAddress);

class KeySpaceHolder {
public:
    uint8_t* data;
    KeySpaceHolder* previousSpace;
    KeySpaceHolder(KeySpaceHolder* existingPreviousSpace, uint8_t* existingData);
    KeySpaceHolder();
    ~KeySpaceHolder();
};
class KeySpace : public KeySpaceHolder {
public:
    int position;
    int size;
    bool fixedSize;
    uint8_t* getTarget();
    KeySpace(bool fixed);
};

#ifndef thread_local
#ifdef __GNUC__
# define thread_local __thread
#elif __STDC_VERSION__ >= 201112L
# define thread_local _Thread_local
#elif defined(_MSC_VER)
# define thread_local __declspec(thread)
#else
# define thread_local
#endif
#endif

// markers for special cases
const double ANY_VERSION = -3.3434325325532E-199;
const double NO_EXIST_VERSION = -4.2434325325532E-199;

KeySpace* getFixedKeySpace();
void setLastVersion(double version);

bool valToBinaryFast(MDB_val &data);
Local<Value> valToUtf8(MDB_val &data);
Local<Value> valToString(MDB_val &data);
Local<Value> valToStringUnsafe(MDB_val &data);
Local<Value> valToBinary(MDB_val &data);
Local<Value> valToBinaryUnsafe(MDB_val &data);
Local<Value> valToNumber(MDB_val &data);
Local<Value> valToBoolean(MDB_val &data);

Local<Value> MDBKeyToValue(MDB_val &data);
void makeGlobalUnsafeBuffer(size_t size);

int putWithVersion(MDB_txn *   txn,
        MDB_dbi     dbi,
        MDB_val *   key,
        MDB_val *   data,
        unsigned int    flags, double version);

void throwLmdbError(int rc);

class TxnWrap;
class DbiWrap;
class EnvWrap;
class CursorWrap;
struct env_path_t {
    MDB_env* env;
    char* path;
    int count;
};

struct action_t {
    int actionType;
    MDB_val key;
    union {
        struct {
            DbiWrap* dw;
        };
        struct {
            MDB_val data;
            double ifVersion;
            double version;
            argtokey_callback_t freeValue;
        };
    };
};

const int INTERRUPT_BATCH = 9998;
const int RESUME_BATCH = 9999;

class BatchWorkerBase : public Nan::AsyncProgressWorker {
  public:
    BatchWorkerBase(Nan::Callback *callback, EnvWrap* envForTxn);
    void ContinueBatch(int rc, bool hasStarted);
    uv_mutex_t* userCallbackLock;
    uv_cond_t* userCallbackCond;
    int interruptionStatus;
    bool finishedProgress;
    EnvWrap* envForTxn;
    TxnWrap* currentTxnWrap;
    ~BatchWorkerBase();
};

/*
    `Env`
    Represents a database environment.
    (Wrapper for `MDB_env`)
*/
class EnvWrap : public Nan::ObjectWrap {
private:
    // List of open read transactions
    std::vector<TxnWrap*> readTxns;
    // Constructor for TxnWrap
    static thread_local Nan::Persistent<Function>* txnCtor;
    // Constructor for DbiWrap
    static thread_local Nan::Persistent<Function>* dbiCtor;
    static uv_mutex_t* envsLock;
    static std::vector<env_path_t> envs;
    static uv_mutex_t* initMutex();
    // compression settings and space
    Compression *compression;
    BatchWorkerBase* batchWorker;

    // Cleans up stray transactions
    void cleanupStrayTxns();

    friend class TxnWrap;
    friend class DbiWrap;

public:
    EnvWrap();
    ~EnvWrap();
    // The wrapped object
    MDB_env *env;
    // Current write transaction
    TxnWrap *currentWriteTxn;

    MDB_txn* currentReadTxn;
    bool readTxnRenewed;
    // Current raw batch transaction
    MDB_txn *currentBatchTxn;
    char* keyBuffer;
    MDB_txn* getReadTxn();

    // Sets up exports for the Env constructor
    static void setupExports(Local<Object> exports);
    void closeEnv();

    /*
        Constructor of the database environment. You need to `open()` it before you can use it.
        (Wrapper for `mdb_env_create`)
    */
    static NAN_METHOD(ctor);
    
    /*
        Gets statistics about the database environment.
    */
    static NAN_METHOD(stat);

    /*
        Gets statistics about the free space database
    */
    static NAN_METHOD(freeStat);
    
    /*
        Detaches a buffer from the backing store
    */
    static NAN_METHOD(detachBuffer);

    /*
        Gets information about the database environment.
    */
    static NAN_METHOD(info);
    /*
        Check for stale readers
    */
    static NAN_METHOD(readerCheck);
    /*
        Print a list of readers
    */
    static NAN_METHOD(readerList);

    /*
        Opens the database environment with the specified options. The options will be used to configure the environment before opening it.
        (Wrapper for `mdb_env_open`)

        Parameters:

        * Options object that contains possible configuration options.

        Possible options are:

        * maxDbs: the maximum number of named databases you can have in the environment (default is 1)
        * maxReaders: the maximum number of concurrent readers of the environment (default is 126)
        * mapSize: maximal size of the memory map (the full environment) in bytes (default is 10485760 bytes)
        * path: path to the database environment
    */
    static NAN_METHOD(open);

    /*
        Resizes the maximal size of the memory map. It may be called if no transactions are active in this process.
        (Wrapper for `mdb_env_set_mapsize`)

        Parameters:

        * maximal size of the memory map (the full environment) in bytes (default is 10485760 bytes)
    */
    static NAN_METHOD(resize);

    /*
        Copies the database environment to a file.
        (Wrapper for `mdb_env_copy2`)

        Parameters:

        * path - Path to the target file
        * compact (optional) - Copy using compact setting
        * callback - Callback when finished (this is performed asynchronously)
    */
    static NAN_METHOD(copy);    

    /*
        Closes the database environment.
        (Wrapper for `mdb_env_close`)
    */
    static NAN_METHOD(close);

    /*
        Starts a new transaction in the environment.
        (Wrapper for `mdb_txn_begin`)

        Parameters:

        * Options object that contains possible configuration options.

        Possible options are:

        * readOnly: if true, the transaction is read-only
    */
    static NAN_METHOD(beginTxn);

    /*
        Opens a database in the environment.
        (Wrapper for `mdb_dbi_open`)

        Parameters:

        * Options object that contains possible configuration options.

        Possible options are:

        * name: the name of the database (or null to use the unnamed database)
        * create: if true, the database will be created if it doesn't exist
        * keyIsUint32: if true, keys are treated as 32-bit unsigned integers
        * dupSort: if true, the database can hold multiple items with the same key
        * reverseKey: keys are strings to be compared in reverse order
        * dupFixed: if dupSort is true, indicates that the data items are all the same size
        * integerDup: duplicate data items are also integers, and should be sorted as such
        * reverseDup: duplicate data items should be compared as strings in reverse order
    */
    static NAN_METHOD(openDbi);

    /*
        Flushes all data to the disk asynchronously.
        (Asynchronous wrapper for `mdb_env_sync`)

        Parameters:

        * Callback to be executed after the sync is complete.
    */
    static NAN_METHOD(sync);

    /*
        Performs a set of operations asynchronously, automatically wrapping it in its own transaction

        Parameters:

        * Callback to be executed after the sync is complete.
    */
    static NAN_METHOD(batchWrite);

    static NAN_METHOD(continueBatch);
    static NAN_METHOD(resetCurrentReadTxn);
};

/*
    `Txn`
    Represents a transaction running on a database environment.
    (Wrapper for `MDB_txn`)
*/
class TxnWrap : public Nan::ObjectWrap {
private:

    // Reference to the MDB_env of the wrapped MDB_txn
    MDB_env *env;

    // Environment wrapper of the current transaction
    EnvWrap *ew;
    // parent TW, if it is exists
    TxnWrap *parentTw;
    // The wrapped object
    MDB_txn *txn;
    
    // Flags used with mdb_txn_begin
    unsigned int flags;

    friend class CursorWrap;
    friend class DbiWrap;
    friend class EnvWrap;

public:
    TxnWrap(MDB_env *env, MDB_txn *txn);
    ~TxnWrap();
    // Remove the current TxnWrap from its EnvWrap
    void removeFromEnvWrap();

    // Constructor (not exposed)
    static NAN_METHOD(ctor);

    // Helper for all the get methods (not exposed)
    static Nan::NAN_METHOD_RETURN_TYPE getCommon(Nan::NAN_METHOD_ARGS_TYPE info, Local<Value> (*successFunc)(MDB_val&));

    // Helper for all the put methods (not exposed)
    static Nan::NAN_METHOD_RETURN_TYPE putCommon(Nan::NAN_METHOD_ARGS_TYPE info, void (*fillFunc)(Nan::NAN_METHOD_ARGS_TYPE info, MDB_val&), void (*freeFunc)(MDB_val&));

    /*
        Commits the transaction.
        (Wrapper for `mdb_txn_commit`)
    */
    static NAN_METHOD(commit);

    /*
        Aborts the transaction.
        (Wrapper for `mdb_txn_abort`)
    */
    static NAN_METHOD(abort);

    /*
        Aborts a read-only transaction but makes it renewable with `renew`.
        (Wrapper for `mdb_txn_reset`)
    */
    static NAN_METHOD(reset);

    /*
        Renews a read-only transaction after it has been reset.
        (Wrapper for `mdb_txn_renew`)
    */
    static NAN_METHOD(renew);

    /*
        Gets string data (JavaScript string type) associated with the given key from a database as UTF-8. You need to open a database in the environment to use this.
        This method is not zero-copy and the return value will usable as long as there is a reference to it.
        (Wrapper for `mdb_get`)

        Parameters:

        * database instance created with calling `openDbi()` on an `Env` instance
        * key for which the value is retrieved
    */
    static NAN_METHOD(getUtf8);

    /*
        Gets string data (JavaScript string type) associated with the given key from a database. You need to open a database in the environment to use this.
        This method is not zero-copy and the return value will usable as long as there is a reference to it.
        (Wrapper for `mdb_get`)

        Parameters:

        * database instance created with calling `openDbi()` on an `Env` instance
        * key for which the value is retrieved
    */
    static NAN_METHOD(getString);

    /*
        Gets string data (JavaScript string type) associated with the given key from a database. You need to open a database in the environment to use this.
        This method is zero-copy and the return value can only be used until the next put operation or until the transaction is committed or aborted.
        (Wrapper for `mdb_get`)

        Parameters:

        * database instance created with calling `openDbi()` on an `Env` instance
        * key for which the value is retrieved
    */
    static NAN_METHOD(getStringUnsafe);

    /*
        Gets binary data (Node.js Buffer) associated with the given key from a database. You need to open a database in the environment to use this.
        This method is not zero-copy and the return value will usable as long as there is a reference to it.
        (Wrapper for `mdb_get`)

        Parameters:

        * database instance created with calling `openDbi()` on an `Env` instance
        * key for which the value is retrieved
    */
    static NAN_METHOD(getBinary);

    /*
        Gets binary data (Node.js Buffer) associated with the given key from a database. You need to open a database in the environment to use this.
        This method is zero-copy and the return value can only be used until the next put operation or until the transaction is committed or aborted.
        (Wrapper for `mdb_get`)

        Parameters:

        * database instance created with calling `openDbi()` on an `Env` instance
        * key for which the value is retrieved
    */
    static NAN_METHOD(getBinaryUnsafe);

    /*
        Gets number data (JavaScript number type) associated with the given key from a database. You need to open a database in the environment to use this.
        This method will copy the value out of the database.
        (Wrapper for `mdb_get`)

        Parameters:

        * database instance created with calling `openDbi()` on an `Env` instance
        * key for which the value is retrieved
    */
    static NAN_METHOD(getNumber);

    /*
        Gets boolean data (JavaScript boolean type) associated with the given key from a database. You need to open a database in the environment to use this.
        This method will copy the value out of the database.
        (Wrapper for `mdb_get`)

        Parameters:

        * database instance created with calling `openDbi()` on an `Env` instance
        * key for which the value is retrieved
    */
    static NAN_METHOD(getBoolean);

    /*
        Puts string data (JavaScript string type) into a database.
        (Wrapper for `mdb_put`)

        Parameters:

        * database instance created with calling `openDbi()` on an `Env` instance
        * key for which the value is stored
        * data to store for the given key
    */
    static NAN_METHOD(putString);

    /*
        Puts string data (JavaScript string type) into a database as UTF-8.
        (Wrapper for `mdb_put`)

        Parameters:

        * database instance created with calling `openDbi()` on an `Env` instance
        * key for which the value is stored
        * data to store for the given key
    */
    static NAN_METHOD(putUtf8);

    /*
        Puts binary data (Node.js Buffer) into a database.
        (Wrapper for `mdb_put`)

        Parameters:

        * database instance created with calling `openDbi()` on an `Env` instance
        * key for which the value is stored
        * data to store for the given key
    */
    static NAN_METHOD(putBinary);

    /*
        Puts number data (JavaScript number type) into a database.
        (Wrapper for `mdb_put`)

        Parameters:

        * database instance created with calling `openDbi()` on an `Env` instance
        * key for which the value is stored
        * data to store for the given key
    */
    static NAN_METHOD(putNumber);

    /*
        Puts boolean data (JavaScript boolean type) into a database.
        (Wrapper for `mdb_put`)

        Parameters:

        * database instance created with calling `openDbi()` on an `Env` instance
        * key for which the value is stored
        * data to store for the given key
    */
    static NAN_METHOD(putBoolean);

    /*
        Deletes data with the given key from the database.
        (Wrapper for `mdb_del`)

        * database instance created with calling `openDbi()` on an `Env` instance
        * key for which the value is stored
    */
    static NAN_METHOD(del);
};

/*
    `Dbi`
    Represents a database instance in an environment.
    (Wrapper for `MDB_dbi`)
*/
class DbiWrap : public Nan::ObjectWrap {
public:
    // Tells how keys should be treated
    NodeLmdbKeyType keyType;
    // Stores flags set when opened
    int flags;
    // The wrapped object
    MDB_dbi dbi;
    // Reference to the MDB_env of the wrapped MDB_dbi
    MDB_env *env;
    // The EnvWrap object of the current Dbi
    EnvWrap *ew;
    // Whether the Dbi was opened successfully
    bool isOpen;
    // compression settings and space
    Compression* compression;
    // versions stored in data
    bool hasVersions;
    // current unsafe buffer for this db
    char* lastUnsafePtr;
    bool getFast;
    bool keysUse32LE;
    bool valuesUse32LE;
    void setUnsafeBuffer(char* unsafePtr, const Persistent<Object> &unsafeBuffer);

    friend class TxnWrap;
    friend class CursorWrap;
    friend class EnvWrap;

    DbiWrap(MDB_env *env, MDB_dbi dbi);
    ~DbiWrap();

    // Constructor (not exposed)
    static NAN_METHOD(ctor);

    /*
        Closes the database instance.
        Wrapper for `mdb_dbi_close`)
    */
    static NAN_METHOD(close);

    /*
        Drops the database instance, either deleting it completely (default) or just freeing its pages.

        Parameters:

        * Options object that contains possible configuration options.

        Possible options are:

        * justFreePages - indicates that the database pages need to be freed but the database shouldn't be deleted

    */
    static NAN_METHOD(drop);
    static NAN_METHOD(dropAsync);

    static NAN_METHOD(stat);
#if ENABLE_FAST_API && NODE_VERSION_AT_LEAST(16,6,0)
    static uint32_t getByBinaryFast(Local<Object> receiver_obj, uint32_t keySize, FastApiCallbackOptions& options);
#endif
    static void getByBinary(const v8::FunctionCallbackInfo<v8::Value>& info);
    static NAN_METHOD(getByPrimitive);
    static NAN_METHOD(getStringByPrimitive);
};

class Compression : public Nan::ObjectWrap {
public:
    char* dictionary;
    char* decompressBlock;
    char* decompressTarget;
    unsigned int decompressSize;
    unsigned int compressionThreshold;
    Persistent<Object> unsafeBuffer;
    // compression acceleration (defaults to 1)
    int acceleration;
    static thread_local LZ4_stream_t* stream;
    void decompress(MDB_val& data, bool &isValid, bool canAllocate);
    argtokey_callback_t compress(MDB_val* value, argtokey_callback_t freeValue);
    void makeUnsafeBuffer();
    void expand(unsigned int size);
    static NAN_METHOD(ctor);
    Compression();
    ~Compression();
    friend class EnvWrap;
    friend class DbiWrap;
};

/*
    `Cursor`
    Represents a cursor instance that is assigned to a transaction and a database instance
    (Wrapper for `MDB_cursor`)
*/
class CursorWrap : public Nan::ObjectWrap {

private:

    // The wrapped object
    MDB_cursor *cursor;
    // Stores how key is represented
    NodeLmdbKeyType keyType;
    // Key/data pair where the cursor is at, and ending key
    MDB_val key, data, endKey;
    // Free function for the current key
    argtokey_callback_t freeKey;
    MDB_cursor_op iteratingOp;
    int flags;
    DbiWrap *dw;
    TxnWrap *tw;
    
    template<size_t keyIndex, size_t optionsIndex>
    friend argtokey_callback_t cursorArgToKey(CursorWrap* cw, Nan::NAN_METHOD_ARGS_TYPE info, MDB_val &key, bool &keyIsValid);

public:
    CursorWrap(MDB_cursor *cursor);
    ~CursorWrap();

    // Sets up exports for the Cursor constructor
    static void setupExports(Local<Object> exports);

    /*
        Opens a new cursor for the specified transaction and database instance.
        (Wrapper for `mdb_cursor_open`)

        Parameters:

        * Transaction object
        * Database instance object
    */
    static NAN_METHOD(ctor);

    /*
        Closes the cursor.
        (Wrapper for `mdb_cursor_close`)

        Parameters:

        * Transaction object
        * Database instance object
    */
    static NAN_METHOD(close);

    // Helper method for getters (not exposed)
    static Nan::NAN_METHOD_RETURN_TYPE getCommon(
        Nan::NAN_METHOD_ARGS_TYPE info, MDB_cursor_op op,
        argtokey_callback_t (*setKey)(CursorWrap* cw, Nan::NAN_METHOD_ARGS_TYPE info, MDB_val&, bool&),
        void (*setData)(CursorWrap* cw, Nan::NAN_METHOD_ARGS_TYPE info, MDB_val&),
        void (*freeData)(CursorWrap* cw, Nan::NAN_METHOD_ARGS_TYPE info, MDB_val&),
        Local<Value> (*convertFunc)(MDB_val &data));

    // Helper method for getters (not exposed)
    static Nan::NAN_METHOD_RETURN_TYPE getCommon(Nan::NAN_METHOD_ARGS_TYPE info, MDB_cursor_op op);

    /*
        Gets the current key-data pair that the cursor is pointing to. Returns the current key.
        This method is not zero-copy and the return value will usable as long as there is a reference to it.
        (Wrapper for `mdb_cursor_get`)

        Parameters:

        * Callback that accepts the key and value
    */
    static NAN_METHOD(getCurrentUtf8);

    /*
        Gets the current key-data pair that the cursor is pointing to. Returns the current key.
        This method is not zero-copy and the return value will usable as long as there is a reference to it.
        (Wrapper for `mdb_cursor_get`)

        Parameters:

        * Callback that accepts the key and value
    */
    static NAN_METHOD(getCurrentString);

    /*
        Gets the current key-data pair that the cursor is pointing to. Returns the current key.
        This method is zero-copy and the value can only be used until the next put operation or until the transaction is committed or aborted.
        (Wrapper for `mdb_cursor_get`)

        Parameters:

        * Callback that accepts the key and value
    */
    static NAN_METHOD(getCurrentStringUnsafe);

    /*
        Gets the current key-data pair that the cursor is pointing to. Returns the current key.
        (Wrapper for `mdb_cursor_get`)

        Parameters:

        * Callback that accepts the key and value
    */
    static NAN_METHOD(getCurrentBinary);


    /*
        Gets the current key-data pair with zero-copy that the cursor is pointing to. Returns the current key.
        This method is zero-copy and the value can only be used until the next put operation or until the transaction is committed or aborted.
        (Wrapper for `mdb_cursor_get`)

        Parameters:

        * Callback that accepts the key and value
    */
    static NAN_METHOD(getCurrentBinaryUnsafe);

    /*
        Gets the current key-data pair that the cursor is pointing to. Returns the current key.
        (Wrapper for `mdb_cursor_get`)

        Parameters:

        * Callback that accepts the key and value
    */
    static NAN_METHOD(getCurrentNumber);

    /*
        Gets the current key-data pair that the cursor is pointing to.
        (Wrapper for `mdb_cursor_get`)

        Parameters:

        * Callback that accepts the key and value
    */
    static NAN_METHOD(getCurrentBoolean);

    /*
    Is the current cursor a database
    (Wrapper for `mdb_cursor_is_db`)
    */
    static NAN_METHOD(getCurrentIsDatabase);

    /*
        Asks the cursor to go to the first key-data pair in the database.
        (Wrapper for `mdb_cursor_get`)
    */
    static NAN_METHOD(goToFirst);

    /*
        Asks the cursor to go to the last key-data pair in the database.
        (Wrapper for `mdb_cursor_get`)
    */
    static NAN_METHOD(goToLast);

    /*
        Asks the cursor to go to the next key-data pair in the database.
        (Wrapper for `mdb_cursor_get`)
    */
    static NAN_METHOD(goToNext);

    /*
        Asks the cursor to go to the previous key-data pair in the database.
        (Wrapper for `mdb_cursor_get`)
    */
    static NAN_METHOD(goToPrev);

    /*
        Asks the cursor to go to the specified key in the database.
        (Wrapper for `mdb_cursor_get`)
    */
    static NAN_METHOD(goToKey);

    /*
        Asks the cursor to go to the first key greater than or equal to the specified parameter in the database.
        (Wrapper for `mdb_cursor_get`)
    */
    static NAN_METHOD(goToRange);

    /*
        For databases with the dupSort option. Asks the cursor to go to the first occurence of the current key.
        (Wrapper for `mdb_cursor_get`)
    */
    static NAN_METHOD(goToFirstDup);

    /*
        For databases with the dupSort option. Asks the cursor to go to the last occurence of the current key.
        (Wrapper for `mdb_cursor_get`)
    */
    static NAN_METHOD(goToLastDup);

    /*
        For databases with the dupSort option. Asks the cursor to go to the next occurence of the current key.
        (Wrapper for `mdb_cursor_get`)
    */
    static NAN_METHOD(goToNextDup);

    /*
        For databases with the dupSort option. Asks the cursor to go to the previous occurence of the current key.
        (Wrapper for `mdb_cursor_get`)
    */
    static NAN_METHOD(goToPrevDup);

    /*
        Go to the entry for next key.
        (Wrapper for `mdb_cursor_get`)
    */
    static NAN_METHOD(goToNextNoDup);

    /*
        Go to the entry for previous key.
        (Wrapper for `mdb_cursor_get`)
    */
    static NAN_METHOD(goToPrevNoDup);

    /*
        For databases with the dupSort option. Asks the cursor to go to the specified key/data pair.
        (Wrapper for `mdb_cursor_get`)
    */
    static NAN_METHOD(goToDup);

    /*
        For databases with the dupSort option. Asks the cursor to go to the specified key with the first data that is greater than or equal to the specified.
        (Wrapper for `mdb_cursor_get`)
    */
    static NAN_METHOD(goToDupRange);

    /*
        Deletes the key/data pair to which the cursor refers.
        (Wrapper for `mdb_cursor_del`)
    */
    static NAN_METHOD(del);

    int returnEntry(int lastRC, MDB_val &key, MDB_val &data);
#if ENABLE_FAST_API && NODE_VERSION_AT_LEAST(16,6,0)
    static uint32_t positionFast(Local<Object> receiver_obj, uint32_t flags, uint32_t offset, uint32_t keySize, uint64_t endKeyAddress, FastApiCallbackOptions& options);
    static uint32_t iterateFast(Local<Object> receiver_obj, FastApiCallbackOptions& options);
#endif
    static void position(const v8::FunctionCallbackInfo<v8::Value>& info);    
    uint32_t doPosition(uint32_t offset, uint32_t keySize, uint64_t endKeyAddress);
    static void iterate(const v8::FunctionCallbackInfo<v8::Value>& info);    
    static NAN_METHOD(renew);
    //static NAN_METHOD(getStringByPrimitive);
};

void load32LE(MDB_val &val, uint32_t* target);
void make32LE(MDB_val &val);

// External string resource that glues MDB_val and v8::String
class CustomExternalStringResource : public String::ExternalStringResource {
private:
    const uint16_t *d;
    size_t l;

public:
    CustomExternalStringResource(MDB_val *val);
    ~CustomExternalStringResource();

    void Dispose();
    const uint16_t *data() const;
    size_t length() const;

    static void writeTo(Local<String> str, MDB_val *val);
};

class CustomExternalOneByteStringResource : public String::ExternalOneByteStringResource {
private:
    const char *d;
    size_t l;

public:
    CustomExternalOneByteStringResource(MDB_val *val);
    ~CustomExternalOneByteStringResource();

    void Dispose();
    const char *data() const;
    size_t length() const;

};


#endif // NODE_LMDB_H
